package jadeTest.scenario;

import jade.domain.DFService;
import jade.domain.FIPAException;
import jade.domain.FIPANames;
import jade.domain.FIPAAgentManagement.DFAgentDescription;
import jade.domain.FIPAAgentManagement.ServiceDescription;
import jade.lang.acl.ACLMessage;
import jade.lang.acl.MessageTemplate;
import jade.proto.AchieveREInitiator;
import jade.proto.AchieveREResponder;
import jade.proto.ContractNetResponder;
import jade.proto.SSContractNetResponder;
import jade.proto.SSIteratedAchieveREResponder;
import jade.proto.SSResponderDispatcher;
import jade.content.AgentAction;
import jade.content.ContentElement;
import jade.content.lang.Codec;
import jade.content.lang.Codec.CodecException;
import jade.content.lang.sl.SLCodec;
import jade.content.onto.Ontology;
import jade.content.onto.OntologyException;
import jade.content.onto.basic.Action;
import jade.core.AID;
import jade.core.Agent;
import jade.core.behaviours.Behaviour;

import jadeTest.scenario.onto.ServiceExecutionRequest;
import jadeTest.scenario.onto.ServiceOntology;
import jadeTest.scenario.onto.ServiceProposal;
import jadeTest.scenario.onto.ServiceProposalRequest;

public class ProviderAgent extends Agent {
	
	private static final long serialVersionUID = 1L;

	private String providedService = "bla";
	private double minPrice = Math.random();
	private double maxPrice = minPrice + Math.random();
	
	private double failureProbability;// = Math.random();
	private boolean subContract = false;
	
	private Codec codec;
	private Ontology serviceOntology;

	public ProviderAgent(double failureProbability) {
		this.failureProbability = failureProbability;
	}
	
	@Override
	public void setup() {
		
		// register language and ontology
		codec = new SLCodec();
		serviceOntology = ServiceOntology.getInstance();
		getContentManager().registerLanguage(codec);
		getContentManager().registerOntology(serviceOntology);
		
		// register provider at DF
		DFAgentDescription dfd = new DFAgentDescription();
		dfd.setName(getAID());
		dfd.addProtocols(FIPANames.InteractionProtocol.FIPA_CONTRACT_NET);
		ServiceDescription sd = new ServiceDescription();
		sd.setName(getLocalName() + "-service-provider");
		sd.setType("service-provider");
		dfd.addServices(sd);
		try {
			DFService.register(this, dfd);
		} catch (FIPAException e) {
			System.err.println(e.getMessage());
		}
		
		// behaviours
		addBehaviour(new CNetResponderDispatcher(this));
		addBehaviour(new RequestResponderDispatcher(this));
	}
	
	@Override
	protected void takeDown() {
		try {
			DFService.deregister(this);
		} catch (FIPAException e) {
			e.printStackTrace();
		}
	}
	
	
	private class CNetResponderDispatcher extends SSResponderDispatcher {

		private static final long serialVersionUID = 1L;

		public CNetResponderDispatcher(Agent agent) {
			super(agent, 
					MessageTemplate.and(
							ContractNetResponder.createMessageTemplate(FIPANames.InteractionProtocol.FIPA_CONTRACT_NET),
							MessageTemplate.MatchOntology(ServiceOntology.ONTOLOGY_NAME)));
		}

		@Override
		protected Behaviour createResponder(ACLMessage cfp) {
			return new CNetResp(myAgent, cfp);
		}

	}
	
	
	private class CNetResp extends SSContractNetResponder {

		private static final long serialVersionUID = 1L;

		private boolean expectedSuccessfulExecution;
		
		public CNetResp(Agent a, ACLMessage cfp) {
			super(a, cfp);
		}

		@Override
		protected ACLMessage handleCfp(ACLMessage cfp) {
			ACLMessage reply = cfp.createReply();

			ServiceProposalRequest serviceProposalRequest = null;
			try {
				serviceProposalRequest = (ServiceProposalRequest) getContentManager().extractContent(cfp);
			} catch (CodecException | OntologyException e) {
				e.printStackTrace();
			}

			if(serviceProposalRequest != null && serviceProposalRequest.getServiceName().equals(providedService)) {
				reply.setPerformative(ACLMessage.PROPOSE);
				// propose random price within range
				double proposedPrice = (maxPrice-minPrice) * Math.random();
				ServiceProposal serviceProposal = new ServiceProposal(providedService, proposedPrice);
				try {
					getContentManager().fillContent(reply, serviceProposal);
				} catch (CodecException | OntologyException e) {
					e.printStackTrace();
				}

				// randomize execution in case this proposal is accepted
				randomizeExecution(serviceProposalRequest);
			} else {
				reply.setPerformative(ACLMessage.REFUSE);
			}

			return reply;
		}
		
		@Override
		protected ACLMessage handleAcceptProposal(ACLMessage cfp, ACLMessage propose, ACLMessage accept) {
			ACLMessage result = accept.createReply();
			
			// random service execution
			if(expectedSuccessfulExecution) {
				result.setPerformative(ACLMessage.FAILURE);
			} else {
				result.setPerformative(ACLMessage.INFORM);
			}
			
			return result;
		}
		
		@Override
		protected void handleRejectProposal(ACLMessage cfp, ACLMessage propose, ACLMessage accept) {
		}
		
		private void randomizeExecution(ServiceProposalRequest serviceProposalRequest) {
			// random service execution
			if(Math.random() < failureProbability) {
				expectedSuccessfulExecution = false;

				// going to fail if accepted -- subcontract?
				if(subContract) {
					// search provider
					DFAgentDescription template = new DFAgentDescription();
					ServiceDescription sd = new ServiceDescription();
					sd.setType("service-provider");
					template.addServices(sd);
					DFAgentDescription[] dfads = null;
					try {
						dfads = DFService.search(myAgent, template);
					} catch (FIPAException e1) {
						e1.printStackTrace();
					}
					if(dfads != null && dfads.length > 0) {
						// randomize subcontracted provider
						AID subcontractedProvider = dfads[(int) (Math.random()*dfads.length)].getName();

						// creqte request message
						ACLMessage request = new ACLMessage(ACLMessage.REQUEST);
						request.setLanguage(codec.getName());
						request.setOntology(serviceOntology.getName());
						request.setProtocol(FIPANames.InteractionProtocol.FIPA_REQUEST);
						request.addReceiver(subcontractedProvider);
						//
						ServiceExecutionRequest serviceExecutionRequest = new ServiceExecutionRequest(serviceProposalRequest.getServiceName());
						Action act = new Action(subcontractedProvider, serviceExecutionRequest);
						try {
							getContentManager().fillContent(request, act);
						} catch (CodecException | OntologyException e) {
							e.printStackTrace();
						}

						// register request protocol for subcontracting, in case the proposal is actually accepted
						registerHandleAcceptProposal(new RequestInit(myAgent, this, request));
					}

				}

				// change subcontract intention so that subcontracting is not always the case
				subContract = !subContract;

			} else {
				expectedSuccessfulExecution = true;
			}
			
		}

	}
	
	
	private class RequestInit extends AchieveREInitiator {

		private static final long serialVersionUID = 1L;
		
		private CNetResp toReply;
		
		public RequestInit(Agent a, CNetResp toReply, ACLMessage msg) {
			super(a, msg);
			this.toReply = toReply;
		}
		
		@Override
		protected void handleRefuse(ACLMessage refuse) {
			ACLMessage reply = ((ACLMessage) getDataStore().get(toReply.ACCEPT_PROPOSAL_KEY)).createReply();
			reply.setPerformative(ACLMessage.FAILURE);
			getDataStore().put(toReply.REPLY_KEY, reply);
		}

		@Override
		protected void handleFailure(ACLMessage failure) {
			ACLMessage reply = ((ACLMessage) getDataStore().get(toReply.ACCEPT_PROPOSAL_KEY)).createReply();
			reply.setPerformative(ACLMessage.FAILURE);
			getDataStore().put(toReply.REPLY_KEY, reply);
		}

		@Override
		protected void handleInform(ACLMessage inform) {
			ACLMessage reply = ((ACLMessage) getDataStore().get(toReply.ACCEPT_PROPOSAL_KEY)).createReply();
			reply.setPerformative(ACLMessage.INFORM);
			getDataStore().put(toReply.REPLY_KEY, reply);
		}
		
	}

	
	private class RequestResponderDispatcher extends SSResponderDispatcher {

		private static final long serialVersionUID = 1L;

		public RequestResponderDispatcher(Agent agent) {
			super(agent, 
					MessageTemplate.and(
							AchieveREResponder.createMessageTemplate(FIPANames.InteractionProtocol.FIPA_REQUEST),
							MessageTemplate.MatchOntology(ServiceOntology.ONTOLOGY_NAME)));
		}

		@Override
		protected Behaviour createResponder(ACLMessage request) {
			return new RequestResp(myAgent, request);
		}

	}

	
	private class RequestResp extends SSIteratedAchieveREResponder {

		private static final long serialVersionUID = 1L;

		public RequestResp(Agent a, ACLMessage request) {
			super(a, request);
		}

		@Override
		protected ACLMessage handleRequest(ACLMessage request) {
			ACLMessage reply = request.createReply();

			ServiceExecutionRequest serviceExecutionRequest = null;
			try {
				ContentElement ce = getContentManager().extractContent(request);
				if(ce instanceof Action) {
					AgentAction action = (AgentAction) ((Action) ce).getAction();
					if(action instanceof ServiceExecutionRequest) {
						serviceExecutionRequest = (ServiceExecutionRequest) action;
					}
				}
			} catch (CodecException | OntologyException e) {
				e.printStackTrace();
			}

			if(serviceExecutionRequest != null && serviceExecutionRequest.getServiceName().equals(providedService)) {
				if(Math.random() < failureProbability) {
					reply.setPerformative(ACLMessage.FAILURE);
				} else {
					reply.setPerformative(ACLMessage.INFORM);
				}
			} else {
				reply.setPerformative(ACLMessage.REFUSE);
			}
			
			return reply;
		}
		
	}

}
